@@ -1,8 +1,11 @@
 mod args;
 mod compile;
+mod completions;
+mod deps;
 mod download;
 mod fonts;
 mod greet;
+mod info;
 mod init;
 mod package;
 mod query;
@@ -20,13 +23,15 @@ use std::io::{self, Write};
 use std::process::ExitCode;
 use std::sync::LazyLock;
 
-use clap::error::ErrorKind;
 use clap::Parser;
+use clap::error::ErrorKind;
 use codespan_reporting::term;
 use codespan_reporting::term::termcolor::WriteColor;
-use typst::diag::HintedStrResult;
+use ecow::eco_format;
+use serde::Serialize;
+use typst::diag::{HintedStrResult, StrResult};
 
-use crate::args::{CliArguments, Command};
+use crate::args::{CliArguments, Command, SerializationFormat};
 use crate::timings::Timer;
 
 thread_local! {
@@ -55,6 +60,9 @@ fn main() -> ExitCode {
     if let Err(msg) = res {
         set_failed();
         print_error(msg.message()).expect("failed to print error");
+        for hint in msg.hints() {
+            print_hint(hint).expect("failed to print hint");
+        }
     }
 
     EXIT.with(|cell| cell.get())
@@ -71,6 +79,8 @@ fn dispatch() -> HintedStrResult<()> {
         Command::Query(command) => crate::query::query(command)?,
         Command::Fonts(command) => crate::fonts::fonts(command),
         Command::Update(command) => crate::update::update(command)?,
+        Command::Completions(command) => crate::completions::completions(command),
+        Command::Info(command) => crate::info::info(command)?,
     }
 
     Ok(())
@@ -86,6 +96,11 @@ fn typst_version() -> &'static str {
     env!("TYPST_VERSION")
 }
 
+/// Used by `args.rs`.
+fn typst_commit_sha() -> &'static str {
+    env!("TYPST_COMMIT_SHA")
+}
+
 /// Print an application-level error (independent from a source file).
 fn print_error(msg: &str) -> io::Result<()> {
     let styles = term::Styles::default();
@@ -98,9 +113,42 @@ fn print_error(msg: &str) -> io::Result<()> {
     writeln!(output, ": {msg}")
 }
 
+/// Print an application-level hint (independent from a source file).
+fn print_hint(msg: &str) -> io::Result<()> {
+    let styles = term::Styles::default();
+
+    let mut output = terminal::out();
+    output.set_color(&styles.header_help)?;
+    write!(output, "hint")?;
+
+    output.reset()?;
+    writeln!(output, ": {msg}")
+}
+
+/// Serialize data to the output format and convert the error to an
+/// [`EcoString`].
+fn serialize(
+    data: &impl Serialize,
+    format: SerializationFormat,
+    pretty: bool,
+) -> StrResult<String> {
+    match format {
+        SerializationFormat::Json => {
+            if pretty {
+                serde_json::to_string_pretty(data).map_err(|e| eco_format!("{e}"))
+            } else {
+                serde_json::to_string(data).map_err(|e| eco_format!("{e}"))
+            }
+        }
+        SerializationFormat::Yaml => {
+            serde_yaml::to_string(data).map_err(|e| eco_format!("{e}"))
+        }
+    }
+}
+
 #[cfg(not(feature = "self-update"))]
 mod update {
-    use typst::diag::{bail, StrResult};
+    use typst::diag::{StrResult, bail};
 
     use crate::args::UpdateCommand;
 
